import win32.windows;
import win32.windef;
import win32.commctrl;

import core.thread;
import std.stdio;
import std.string;
import std.utf;

ushort LOWORD(uint l){ return l & 0xffff; }
ushort HIWORD(uint l){ return l >>> 16; }
COLORREF RGBA(uint r, uint g, uint b, uint a) {
  return cast(COLORREF)(r | (g << 8) | (b << 16) | (a<<24));
}

shared static ProgressThread[int] _ProgressThreadList ;

enum CMD{
  INVAL = 0x900
} ;

class ProgressWindow {
  ProgressThread ThProgressWindow ;
  string classname ;
  string name ;
  
  this(string name){
    this.name=name ;
  }
  
  int start(int x, int y, int w, int h){
    ThProgressWindow = new ProgressThread(name) ;
    ThProgressWindow.setPosition(x,y,w,h) ;
    ThProgressWindow.start() ;
    return(0) ;
  }

  int setRate(int val){
    ThProgressWindow.setRate(val) ;
    return(0) ;
  }
  
  int setValue(int val1, int val2){
    ThProgressWindow.setValue(val1,val2) ;
    return(0) ;
  }
  
  int close(){
    // WM_DESTROYメッセージを送った時に ウインドタイトルバーをドラッグしてると
    // メッセージが消失するらしい。スレッドがrunning中は WM_DESTROYを再送し続ける
    // コードにした。
    while(1){
      if( ThProgressWindow.isRunning()==false ) break ;
      SendMessage(ThProgressWindow.hwnd, WM_DESTROY, 0,0) ;
    }
    ThProgressWindow.join() ;
    return(0) ;
  }
}

class  ProgressThread : Thread{
  HWND        hwnd ;
  WNDCLASSEX  winc ;
  string      winname ;
  int         exstyle ;
  int         style   ;
  int         winX,winY,winW,winH ;
  int         rate ;
  int         val1, val2 ;
  int         barx, bary ;
  int         barheight ;
  int         barwidth  ;

  HFONT       hFont;
  
  this(string name){
    this.winc.cbSize        = WNDCLASSEX.sizeof;
    this.winc.lpszClassName = cast(wchar*)std.utf.toUTF16z("PROGRESS") ;
    this.winc.hInstance     = GetModuleHandle(null) ;
    this.winc.lpfnWndProc   = cast(WNDPROC)(&_WndProgressProc) ;
    //this.winc.hbrBackground = CreateSolidBrush(GetSysColor(COLOR_BTNFACE)) ;
    this.hwnd               = null ;
    this.winname            = name ;
    
    RegisterClassEx(cast(WNDCLASSEXW*)(&this.winc)) ;

    style = WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_CAPTION | WS_OVERLAPPED | WS_VISIBLE ;

    super( &run ) ;
  }

  ~this(){
    _ProgressThreadList.remove(cast(int)(this.hwnd)) ;
  }

  void run(){
    int  bRet ;
    MSG  msg ;
    
    Thread[] thlist ;

    hwnd = CreateWindowEx(exstyle,
			  cast(wchar*)this.winc.lpszClassName,
			  std.utf.toUTF16z(this.winname),
			  style,
			  winX,winY,winW,winH,
			  null, null, this.winc.hInstance , cast(LPVOID)(this)
			  );
    
    while( (bRet = GetMessage(&msg , null , 0 , 0)) != 0 ){
      if (bRet == -1 ) break;
      DispatchMessage(&msg);
    }
  }

  int setPosition(int x, int y, int w, int h){
    this.winX=x ;
    this.winY=y ;
    this.winW=w ;
    this.winH=h ;
    return(0) ;
  }

  int setRate(int rate){
    synchronized{
      this.rate=rate ;
      this.val1=rate ;
      this.val2=   1 ;
      SendMessage(hwnd,WM_COMMAND,CMD.INVAL,0) ;
    }
    return(0) ;
  }

  int setValue(int val1, int val2){
    synchronized{
      this.val1=val1 ;
      this.val2=val2 ;
      this.rate=val1*100/val2 ;
      SendMessage(hwnd,WM_COMMAND,CMD.INVAL,0) ;
    }
    return(0) ;
  }

  LRESULT proc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    switch (msg) {
    case WM_DESTROY:
      DeleteObject(hFont);
      PostQuitMessage(0);
      break ;
      
    case WM_CREATE :
      RECT rc ;
      GetClientRect( hwnd, &rc ) ;
      winW=rc.right  ;
      winH=rc.bottom ;

      barheight = 20 ;
      barwidth  = winW*4/6 ;
      barx      = winW/6 ;
      bary      = winH/2 - barheight/2 ;
      hFont = CreateFont(
			 14 , 0 , 0 , 0 , FW_REGULAR , FALSE , FALSE , FALSE ,
			 SHIFTJIS_CHARSET , OUT_DEFAULT_PRECIS ,
			 CLIP_DEFAULT_PRECIS , DEFAULT_QUALITY , 
			 VARIABLE_PITCH, null
			 );
      break ;

    case WM_PAINT:
      PAINTSTRUCT ps;
      HDC         hdc;
      HPEN        PPRG, PRST, PBG ;
      HBRUSH      BPRG, BRST, BBG ;

      PPRG = CreatePen(PS_SOLID, 1, RGBA(  0,255,  0,0) & 0x00ffffff) ;
      PRST = CreatePen(PS_SOLID, 1, RGBA(128,128,128,0) & 0x00ffffff) ;
      PBG  = CreatePen(PS_SOLID, 1, GetSysColor(COLOR_BTNFACE)) ;
      BPRG = CreateSolidBrush(RGBA(  0,255,  0,0) & 0x00ffffff) ;
      BRST = CreateSolidBrush(RGBA(128,128,128,0) & 0x00ffffff) ;
      BBG  = CreateSolidBrush(GetSysColor(COLOR_BTNFACE)) ;
      
      hdc = BeginPaint(hwnd, &ps);
      SetBkMode(hdc,TRANSPARENT) ;
      int barsx=barx ;
      int barsy=bary ;
      int barex=barx+barwidth*rate/100 ;
      int barey=bary+barheight ;
      int rstsx=barx+barwidth*rate/100 ;
      int rstsy=bary ;
      int rstex=barx+barwidth ;
      int rstey=bary+barheight ;
      
      SelectObject(hdc, PBG) ;
      SelectObject(hdc, BBG) ;
      Rectangle(hdc,     0,    0,barsx,winH ) ;
      Rectangle(hdc, rstex,    0, winW,winH ) ;
      Rectangle(hdc, barsx,    0,rstex,barsy) ;
      Rectangle(hdc, barsx,barey,rstex,winH ) ;
      
      SelectObject(hdc, PPRG) ;
      SelectObject(hdc, BPRG) ;
      Rectangle(hdc, barsx,barsy,barex,barey) ;
      SelectObject(hdc, PRST) ;
      SelectObject(hdc, BRST) ;
      Rectangle(hdc, rstsx,rstsy,rstex,rstey) ;

      SelectObject(hdc , hFont);

      string str ;
      synchronized{
	str = format("%3d%%",rate) ;
      }
      TextOut(hdc, barx+barwidth/2, bary+2, cast(wchar*)(toUTF16(str)) , str.length);

      EndPaint(hwnd, &ps);
      DeleteObject(PPRG) ;
      DeleteObject(PRST) ;
      DeleteObject(PBG ) ;
      DeleteObject(BPRG) ;
      DeleteObject(BRST) ;
      DeleteObject(BBG ) ;
      break ;
      
    case WM_CLOSE:
      DestroyWindow(hwnd);
      break ;

    case WM_COMMAND:
      switch( LOWORD(wp) ){
      case CMD.INVAL :
	InvalidateRect(hwnd , null , true) ;
	break ;
      default :	break ;
      }
      break ;
      
    default :
      return( DefWindowProc(hwnd , msg , wp , lp) ) ;
    }
    return(0) ;
  }
}

extern (Windows) LRESULT _WndProgressProc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp)
{
  if( msg==WM_CREATE ){
    synchronized{
      ProgressThread pt = cast(ProgressThread)(*(cast(int*)(lp))) ;
      pt.hwnd=hwnd ;
      _ProgressThreadList[cast(int)(hwnd)]=cast(shared(ProgressThread))(pt) ;
    }
  }
  if( (cast(int)(hwnd) in _ProgressThreadList) != null ){
    ProgressThread pt = cast(ProgressThread)(_ProgressThreadList[cast(int)(hwnd)]) ;
    LRESULT res = pt.proc(hwnd,msg,wp,lp) ;
    return( res ) ;
  }
  return( DefWindowProc(hwnd , msg , wp , lp) ) ;
}
