import win32.windows;
import win32.windef;
import win32.commctrl;
import win32frame.WindowObject;
import std.stdio;
import std.string;
import std.utf;

ushort LOWORD(uint l){ return l & 0xffff; }
ushort HIWORD(uint l){ return l >>> 16; }
COLORREF RGBA(uint r, uint g, uint b, uint a) {
  return cast(COLORREF)(r | (g << 8) | (b << 16) | (a<<24));
}

const LBUTTONDOWN   = (0x0001) ;
const LBUTTONUP     = (0x0002) ;
const LBUTTONDBLCLK = (0x0004) ;
const RBUTTONDOWN   = (0x0001) ;
const RBUTTONUP     = (0x0002) ;
const RBUTTONDBLCLK = (0x0004) ;
const ALT_KEY       = (0x0001) ;
const CNTL_KEY      = (0x0002) ;
const SHIFT_KEY     = (0x0004) ;

class SliderBase : WindowObject {
  int mouseX  =0 ;
  int mouseY  =0 ;
  int mouseB  =0 ;
  int ACS_stat=0 ;
  double min, max, cur ;
  int sliderx, slidery ;
  int slidertype ;
  int touching ;
  COLORREF colfg, colbg ;
  

  this(string name, string ClassName) {
    winc.style          = 0; //CS_HREDRAW | CS_VREDRAW ;
    winc.cbClsExtra     = 0;
    winc.cbWndExtra     = 0;
    winc.lpszMenuName   = null;
    winc.hCursor        = LoadCursor(null , cast(wchar*)(IDC_ARROW));
    //    winc.hbrBackground  = cast(HBRUSH)(GetStockObject(BLACK_BRUSH));
    winc.lpszClassName  = cast(wchar*)(ClassName) ;
    super(name) ;
    this.registwinc() ;
    slidertype=0 ;
    min=0.0 ;
    max=1.0 ;
    cur=0.5 ;
    colfg=RGBA(128,128,128,255) ;
    colbg=RGBA(128,128,128,255) ;
  }

  int setInfo(int slidertype, double min, double max, double cur){
    if( min<max && (min<=cur) && (cur<=max) ){
      this.min=min ;
      this.max=max ;
      this.cur=cur ;
      this.slidertype=slidertype ;
      return(0) ;
    }
    return(-1) ;
  }
  
  int setValue( double cur ){
    if( (min<=cur) && (cur<=max) ){
      this.cur=cur ;
      if( hwnd !is null ){
	sliderx = cast(int)(cast(double)(winW-1) * cur/(max-min)) ;
	slidery = cast(int)(cast(double)(winH-1) * cur/(max-min)) ;
	RedrawWindow( hwnd, null, null, RDW_INVALIDATE) ;
      }
      return(0) ;
    }
    return(-1) ;
  }

  int setFGcolor(uint r, uint g, uint b, uint a){
    colfg=RGBA(r,g,b,a) ;
    return(0) ;
  }
  
  int setBGcolor(uint r, uint g, uint b, uint a){
    colbg=RGBA(r,g,b,a) ;
    return(0) ;
  }
    
    
  double getValue(){
    if( slidertype==0 ){
      cur=(max-min) * cast(double)(sliderx)/(cast(double)(winW-1)) ;
    }else{
      cur=(max-min) * cast(double)(slidery)/(cast(double)(winH-1)) ;
    }
    return(cur) ;
  }

  LRESULT proc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    switch (msg) {
    case WM_DESTROY:
      this.wm_destroy(hwnd, wp ,lp) ;
      break ;
      
    case WM_CREATE:
      this.wm_create(hwnd, wp ,lp) ;
      break ;
      
    case WM_PAINT:
      this.wm_paint(hwnd, wp, lp) ;
      if( touching==1 ){
	sendmsgtoparent( WM_COMMAND, cast(int)(this.hmenu), 0) ;
      }
      break ;

    case WM_MOVE:
      sliderx = cast(int)(cast(double)(winW-1) * cur/(max-min)) ;
      slidery = cast(int)(cast(double)(winH-1) * cur/(max-min)) ;
      RedrawWindow( hwnd, null, null, RDW_INVALIDATE) ;
      break ;

    case WM_SIZE:
      winW=LOWORD(lp) ;
      winH=HIWORD(lp) ;
      this.wm_size(winW,winH) ;
      sliderx = cast(int)(cast(double)(winW-1) * cur/(max-min)) ;
      slidery = cast(int)(cast(double)(winH-1) * cur/(max-min)) ;
      RedrawWindow( hwnd, null, null, RDW_INVALIDATE) ;
      break;

    case WM_LBUTTONDOWN:
    case WM_LBUTTONUP:
    case WM_LBUTTONDBLCLK:
    case WM_RBUTTONDOWN:
    case WM_RBUTTONUP:
    case WM_RBUTTONDBLCLK:
    case WM_MOUSEMOVE:
      {
	POINT point;

	GetCursorPos(&point);
	ScreenToClient( hwnd, &point );
	mouseX = point.x ;
	mouseY = point.y ;
	if( msg==WM_LBUTTONDOWN ){
	  touching=1 ;
	  SetFocus(hwnd) ;
	  SetCapture( hwnd ) ;
	}
	if( msg==WM_LBUTTONUP   ){
	  touching=0 ;
	  ReleaseCapture() ;
	}
	if( msg==WM_MOUSEMOVE ){
	}else{
	  mouseB=0 ;
	  switch( msg ){
	  case WM_LBUTTONDOWN:
	    mouseB|=LBUTTONDOWN ;
	    break ;
	  case WM_LBUTTONUP:
	    mouseB|=LBUTTONUP ;
	    break ;
	  case WM_LBUTTONDBLCLK:
	    mouseB|=LBUTTONDBLCLK ;
	    break ;
	  case WM_RBUTTONDOWN:
	    mouseB|=RBUTTONDOWN ;
	    break ;
	  case WM_RBUTTONUP:
	    mouseB|=RBUTTONUP ;
	    break ;
	  case WM_RBUTTONDBLCLK:
	    mouseB|=RBUTTONDBLCLK ;
	    break ;
	  }	  
	  
	  if(GetKeyState(VK_SHIFT)   < 0) ACS_stat |=   SHIFT_KEY  ;
	  else                            ACS_stat &= ~(SHIFT_KEY) ;
	  if(GetKeyState(VK_CONTROL) < 0) ACS_stat |=   CNTL_KEY   ;
	  else                            ACS_stat &= ~(CNTL_KEY)  ;
	  if(GetKeyState(VK_MENU)    < 0) ACS_stat |=   ALT_KEY    ;
	  else                            ACS_stat &= ~(ALT_KEY)   ;
	  
	}
	if( (mouseB & LBUTTONDOWN)!=0 ){
	  sliderx = (mouseX > winW) ? winW-1 : (mouseX<0) ? 0 : mouseX ;
	  slidery = (mouseY > winH) ? winH-1 : (mouseY<0) ? 0 : mouseY ;
	  RedrawWindow( hwnd, null, null, RDW_INVALIDATE) ;
	}
      }
      break ;
      
    default :
      return( DefWindowProc(hwnd , msg , wp , lp) );
    }
    return(0);
  }

  int wm_destroy(HWND hwnd, WPARAM wp , LPARAM lp){
    return(0) ;
  }
  
  int wm_create(HWND hwnd, WPARAM wp , LPARAM lp){
    return(0) ;
  }
  
  int wm_paint(HWND hwnd, WPARAM wp , LPARAM lp){
    HDC hdc;
    PAINTSTRUCT ps ;
    HPEN   PCOLFG, PCOLBG ;
    HBRUSH BCOLFG, BCOLBG ;

    PCOLFG = CreatePen(PS_SOLID, 1, colfg & 0x00ffffff ) ;
    PCOLBG = CreatePen(PS_SOLID, 1, colbg & 0x00ffffff ) ;
    BCOLFG = CreateSolidBrush(colfg & 0x00ffffff) ;
    BCOLBG = CreateSolidBrush(colbg & 0x00ffffff) ;
    
    hdc = BeginPaint(hwnd, &ps);

    if( slidertype==0 ){
      SelectObject(hdc, PCOLBG) ;
      SelectObject(hdc, BCOLBG) ;
      Rectangle(hdc, sliderx,0,winW,winH) ;
      SelectObject(hdc, PCOLFG) ;
      SelectObject(hdc, BCOLFG) ;
      Rectangle(hdc, 0,0,sliderx,winH) ;
    }else{
      SelectObject(hdc, PCOLBG) ;
      SelectObject(hdc, BCOLBG) ;
      Rectangle(hdc, 0,0,winW,slidery) ;
      SelectObject(hdc, PCOLFG) ;
      SelectObject(hdc, BCOLFG) ;
      Rectangle(hdc, 0,winH,winW,slidery) ;
    }
    
    EndPaint(hwnd, &ps);

    DeleteObject(PCOLFG) ;
    DeleteObject(PCOLBG) ;
    DeleteObject(BCOLFG) ;
    DeleteObject(BCOLBG) ;
	
    return(0) ;
  }
  
  int wm_size(int w, int h){
    return(0) ;
  }
  
}
